/*------------------------------------------------------------------------------

  ADRINAS: Automatic Detection and Removal of Impulsive Noise in Audio Signals

  This code is part of the following publication and was subject
  to peer review:

    "Automatic Detection and Removal of Impulsive Noise in Audio Signals"
    by Laurent Oudre, Image Processing On Line, 2015.
    DOI:10.5201/ipol.2015.XXXXX
    http://dx.doi.org/10.5201/ipol.2015.XXXXX

  Copyright (c) 2012-2015 Laurent Oudre <laurent.oudre@cmla.ens-cachan.fr>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU Affero General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>.

------------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include "sound.h"
#include "adrinas.h"

/*----------------------------------------------------------------------------*/
int main(int argc, char ** argv)
{
  double K = 2.0;    /* First threshold */
  int b = 20;        /* Second threshold */
  int p = 3*100+2;   /* Order of the model */
  int Nw = 8*p;      /* Window length */
  Sound * signal = NULL;
  int * burst1 = NULL;
  int * burst2 = NULL;
  int nSamples = 0;
  int i,c;

  /* read input input */
  if( argc < 3 )
    {
      fprintf(stderr,"usage: adrinas in.wav out.wav [K b p Nw]\n"
                     "optional parameters:\n"
                     "       K: threshold parameter    (default value %0.2f)\n"
                     "       b: burst fusion parameter (default value %d)\n"
                     "       p: order of AR model      (default value %d)\n"
                     "       Nw: window size           (default value %d)\n",
                     K,b,p,Nw);
      exit(EXIT_FAILURE);
    }
  signal = audioRead(argv[1],1,20000000);
  if( argc >= 4 ) K  = atof(argv[3]);
  if( argc >= 5 ) b  = atoi(argv[4]);
  if( argc >= 6 ) p  = atoi(argv[5]);
  if( argc >= 7 ) Nw = atoi(argv[6]);
  printf("input signal: %ld channel, %ld samples\n",
         signal->nChannels,signal->nSamples);
  printf("parameters: K=%.2f b=%d p=%d Nw=%d\n\n",K,b,p,Nw);

  /* process each channel */
  for(c=0; c<signal->nChannels; c++)
    {
      printf("processing channel #%d:\n",c);

      /* iteration 1 */
      burst1 = adrinas(signal->channel[0],signal->nSamples,p,K,b,Nw);

      /* iteration 2 */
      burst2 = adrinas(signal->channel[0],signal->nSamples,p,K,b,Nw);

      /* compute number of samples detected as noise */
      for(i=0; i<(signal->nSamples); i++)
        if( (burst1[i]==1) || (burst2[i]==1) ) ++nSamples;
      printf("  samples detected as noise: %d (%0.2f%%)\n",
             nSamples, 100.0 * nSamples / signal->nSamples);
    }

  /* write output */
  audioWrite(signal,argv[2]);

  /* free memory */
  soundClean(signal);
  free(burst1);
  free(burst2);

  return 0;
}
/*----------------------------------------------------------------------------*/
